#include <stddef.h>
#include <string.h>
#include "myfnmtch.h"

static int _toupper (int c)
{
  if ((c >= 'a') && (c <= 'z'))
    c -= 'a' - 'A';
  return c;
}

#define _isslash(c)     ((c) == '\\' || (c) == '/')

#define _findslash(s)   strpbrk ((s), "\\/")

// This function returns pattern if the test character matches it.
// Otherwise it returns NULL.
// nocase specifies whether the comparison is case-insensetive or not.
// pattern must not be NULL.
// Since this is a list/range matching, pattern should point to a (sub)string
// containing ']' for a match.
// pattern should point to the character after [.
static const char* BracketExpMatch (const char* pattern, char test, int nocase)
{
  char c, c2;
  int negate, match = 0, SpecialBracketMeaning = 0;

  if ((negate = (*pattern == '!')) != 0)
    pattern++;

  for (;;)
  {
    // Allow []...] and [!]...] subpatterns:
    if (((c = *pattern++) == ']') && SpecialBracketMeaning)
      break;

    // Subpattern must end in ] for a match:
    if (!c)
      return NULL;

    // Special chars don't work here (?, *, back slash for escaping,
    // any slash for exact slash match)

    SpecialBracketMeaning = 1;

    if ((*pattern == '-') && ((c2 = pattern[1]) != 0) && (c2 != ']'))
    {
      if ((c <= test) && (test <= c2))
        match = 1;
      else if (
               nocase &&
               (_toupper((unsigned char)c) <= _toupper((unsigned char)test)) &&
               (_toupper((unsigned char)test) <= _toupper((unsigned char)c2))
              )
        match = 1;
      pattern += 2;
    }
    else if (c == test)
      match = 1;
    else if (nocase &&
             (_toupper((unsigned char)c) == _toupper((unsigned char)test)))
      match = 1;
  }

  return (match == negate) ? NULL : pattern;
}

int fnmatch (const char* pattern, const char* string, int flags)
{
  char c;
  char test;

  for (;;)
  {
    switch ((c = *pattern++))
    {
    case 0:
      // match if sucessfully arrived at the end of pattern and string.
      // no match otherwise:
      return (!*string) ? 0 : FNM_NOMATCH;
      
    case '?': // matches 1 char
      // no match if string has been fully scanned and there's nothing left
      // to match '?' OR there's a slash to match '?' but it must match a
      // slash exactly:
      if (!(test = *string++) ||
          ((flags & FNM_PATHNAME) && _isslash(test)))
        return FNM_NOMATCH;
      break; // switch ((c = *pattern++))
      
    case '*': // matches 0 or more chars
      // collapse multiple asterisks:
      while ((c = *pattern) == '*')
        pattern++;

      // optimize for pattern ending with *:
      if (!c)
      {
        // no match if there's a slash in string and it requires an exact match
        if ((flags & FNM_PATHNAME) && (_findslash(string) != NULL))
          return FNM_NOMATCH;
        // match otherwise
        return 0;
      }
      // optimize for pattern with * followed by a slash if exact
      // slash match required:
      else if (
                (flags & FNM_PATHNAME) && _isslash(c)
                // $$$ what if it's an escaped char, not just a slash?
                &&
                (
                  (c == '/') ||
                  (flags & FNM_NOESCAPE) ||
                  _isslash(pattern[1]) // escaped slash
                )
              )
      {
        // no match if there's no slash in string:
        if ((string = _findslash(string)) == NULL)
          return FNM_NOMATCH;
        // else skip string until the slash (skip what matches with *)
        // and continue the matching process from there:
        break; // switch ((c = *pattern++))
      }

      // if there's something other than \0 or slash after *,
      // step by step try expanding * onto more and more of string
      // until there's a match of string with remaining pattern
      // (use recursion):
      while ((test = *string) != 0)
      {
        // a match is a match:
        if (!fnmatch(pattern, string, flags))
          return 0;
        // no match if there's a slash in string requiring an exact match:
        if ((flags & FNM_PATHNAME) && _isslash(test))
          break; // while ((test = *string) != 0)
        // expand * onto more of string:
        string++;
      }
      // no match from recursion or no exact slash match:
      return FNM_NOMATCH;
      
    case '[': // bracketed range matches 1 char
      // no match if string has been fully scanned and there's nothing left
      // to match a range OR there's a slash to match the range but the
      // slash must match exactly (in exact slash matching mode a slash
      // in string must not match the bracketed range):
      if (!(test = *string++) ||
          ((flags & FNM_PATHNAME) && _isslash(test)))
        return FNM_NOMATCH;
      if ((pattern = BracketExpMatch(pattern, test, flags & FNM_NOCASE)) == NULL)
        return FNM_NOMATCH;
      break; // switch ((c = *pattern++))
      
    case '\\': // either escaped char or slash
      if (!(flags & FNM_NOESCAPE))
      {
        // it's an escaped character:

        // $$$ escaped \0 should be handled as unescaped \0,
        // and so should be ordinary (non-special) chars:
        if (!(c = *pattern) || (strchr("\\?*[", c) == NULL))
          break; // switch ((c = *pattern++))

        // skip the escaping '\\' in pattern and fall through treating
        // the escaped special chars as ordinary chars:
        pattern++;
      }
      // fall through
      
    default:
      // this special slash matching check is for '/' in pattern
      // and fall-throughs from the above:
      if (_isslash(c) && _isslash(*string))
      {
        string++;
        break;
      }

      if (flags & FNM_NOCASE)
      {
        if (_toupper((unsigned char)c) != _toupper((unsigned char)*string++))
          return FNM_NOMATCH;
      }
      else
      {
        if (c != *string++)
          return FNM_NOMATCH;
      }
      break;
    }
  }
}

#ifdef TEST_FN_MATCH
#include <stdio.h>
int main (int argc, char* argv[])
{
  int flags;

  if (argc != 4)
  {
    printf
    (
      "int fnmatch (const char* pattern, const char* string, int flags)\n"
    );
    return 0;
  }

  if (sscanf (argv[3], "%d", &flags) != 1)
    return 0;

  printf
  (
    "fnmatch(\"%s\",\"%s\",%d)=%d\n",
    argv[1],
    argv[2],
    flags,
    fnmatch (argv[1], argv[2], flags)
  );

  return 0;
}
#endif
