#include <string.h>

#define MAX(a,b)        ( ((a)>(b)) ? (a) : (b) )

/*
 * Splits the given path to the directory and name parts.
 *
 * If a slash exists in the path, then
 * the directory part is the slash and everything before it.
 *
 * If no slash exists but a colon exists (presumably with a disk letter before
 * it), then the directory part is the colon and everything before it.
 *
 * In the case a slash/colon existed, the name part is the rest of the path.
 *
 * If the path is "." or "..", the directory part is the path and
 * the name is an empty string.
 *
 * Otherwise (i.e. neither slash, nor colon, nor "."("..") were found in the
 * path), the directory part is set to "." and the name part is the path.
 *
 * Note 1: if the path is of the form: *\name or ?:name (i.e. w/o a trailing
 * slash), the entire path may specify either a directory or a file.
 * To distinguish between the two, an additional check must be made by
 * the caller.
 *
 * Note 2: if the returned directory part ends with a colon, no slash must
 * be appended to it to form a valid directory/file name.
 * Otherwise, depending on the application, a slash should be appended to the
 * directory part (if it contains no trailing slash) or the trailing slash
 * (if it exists) should be removed.
 *
 * Note 3: the returned name part can be used to match against a pattern
 * (with wildcards such as ?* and perhaps ranges) with e.g. fnmatch() function.
 * It should be noted, however, in DOS a filename w/o a trailing dot
 * matches to a pattern having a trailing dot (e.g. "name" matches "*."),
 * this is a special kind of pattern for finding files w/o filename extension.
 *
 * Note 4: the function does not allocate memory nor checks if the supplied
 * buffers are big enough to fit the directory and name parts.
 *
 * Returned value: 0 on success, -1 on error (if null pointer detected).
 */
int pathsplit (const char* path, char* dir, char* name)
{
  char* pLast;

  if ((path == NULL) || (dir == NULL) || (name == NULL))
    return -1;

  pLast = MAX (strrchr (path, '\\'), strrchr (path, '/'));
  pLast = MAX (pLast, strrchr (path, ':'));

  if (pLast != NULL)
  {
    strcpy (name, pLast+1);
    memcpy (dir, path, pLast+1-(char*)path);
    dir[pLast+1-(char*)path] = 0;
    return 0;
  }

  if ((!strcmp(path,".")) || (!strcmp(path,"..")))
  {
    strcpy (dir, path);
    *name = 0;
    return 0;
  }

  strcpy (dir, ".");
  strcpy (name, path);
  return 0;
}
