#include "alloc.h"

#ifdef MAX
  #undef MAX
#endif
#define MAX(a,b) (((a)>(b))?(a):(b))

#ifdef MIN
  #undef MIN
#endif
#define MIN(a,b) (((a)<(b))?(a):(b))

#ifdef C_ASSERT
  #undef C_ASSERT
#endif
#define C_ASSERT(x) extern char __aSsErT__[1-2*!(x)];

#define MCB_FLAG_ALLOCATED      0x0001
#define MCB_FLAG_LAST           0x0002

#define MCB_HEAD_TAIL_SIZE      (sizeof(tMCBHead) + sizeof(tMCBTail))
#define MIN_MCB_SIZE            MCB_HEAD_TAIL_SIZE
#define MIN_BLOCK_SIZE          (sizeof(int))

// the code needs sizeof(size_t) >= sizeof(void*)
C_ASSERT(sizeof(size_t) >= sizeof(void*))

#define MAPPING_USED    0
#define DONE_USED       0
#define STATISTICS_USED 0

static int map (tAlloc* pAlloc, void* pBuf, size_t size)
{
#if MAPPING_USED
  // will map more mem if buffer isn't page-aligned or
  // its size isn't a multiple of page size
  size_t psm1 = pAlloc->PageSize-1;
  size_t p = (size_t)pBuf & ~psm1;
  size_t diff = (size_t)pBuf - p;
  size_t sz = (size + diff + psm1) & ~psm1;
  return pAlloc->pMapPages (pAlloc, (void*)p, sz / ++psm1);
#else
  (void)pAlloc;
  (void)pBuf;
  (void)size;
  return 1;
#endif
}

static void unmap (tAlloc* pAlloc, void* pBuf, size_t size)
{
#if MAPPING_USED
  // will unmap less mem if buffer isn't page-aligned or
  // its size isn't a multiple of page size
  size_t psm1 = pAlloc->PageSize-1;
  size_t p = ((size_t)pBuf + psm1) & ~psm1;
  size_t diff = p - (size_t)pBuf;
  size_t sz = (size - diff) & ~psm1;
  if (size <= diff)
    return;
  if (sz)
    pAlloc->pUnmapPages (pAlloc, (void*)p, sz / ++psm1);
#else
  (void)pAlloc;
  (void)pBuf;
  (void)size;
#endif
}

int ALLOC_Init (tAlloc* pAlloc, tAllocInit* pAllocInit)
{
  tMCBHead* pHead;
  tMCBTail* pTail;
  size_t size;
  size_t psm1;

  // Check pointers, sizes and other stuff
  if (pAlloc != NULL)
  {
     pAlloc->pHeapEnd = pAlloc->pHeapBegin = NULL;
  }

  if ((pAlloc == NULL) || (pAllocInit == NULL))
    return 0;

  if (pAllocInit->pHeapBegin == NULL)
    return 0;

  pAlloc->PageSize = pAllocInit->PageSize;

  // Prepare the initial (big) block
  if (!pAlloc->PageSize)
  {
    // take heap beginning address and heap size as is
    size = pAllocInit->HeapSize;
    pHead = pAlloc->pHeapBegin = pAllocInit->pHeapBegin;
  }
  else
  {
    size_t p = (size_t)pAllocInit->pHeapBegin;
    size_t diff;
    psm1 = pAlloc->PageSize-1;
    if (pAlloc->PageSize & psm1)
      return 0;                 // error: page size isn't a power of 2
    // align heap beginning address to page size boundary
    p = (p + psm1) & ~psm1;
    diff = p - (size_t)pAllocInit->pHeapBegin;
    size = pAllocInit->HeapSize;
    if (size <= diff)
      return 0;
    // round heap size to page size
    size = (size - diff) & ~psm1;
    pHead = pAlloc->pHeapBegin = (void*)p;
    pAlloc->pMapPages = pAllocInit->pMapPages;
    pAlloc->pUnmapPages = pAllocInit->pUnmapPages;
  }

  if (size < MIN_MCB_SIZE)
    return 0;                   // error: too little space for a single block

  pAlloc->pHeapEnd = (void*)((char*)pAlloc->pHeapBegin + size);
  pTail = (void*)((char*)pAlloc->pHeapEnd - sizeof(tMCBTail));

  // Allocate page frames for head and tail tags
  if (pAlloc->PageSize)
    if
    (
      (!map (pAlloc, pHead, sizeof(tMCBHead))) ||
      (!map (pAlloc, pTail, sizeof(tMCBTail)))
    )
    {
      unmap (pAlloc, pHead, size);
      return 0;                 // error: failed to map in page frames
    }

  pTail->size = pHead->size = size - MCB_HEAD_TAIL_SIZE;

  // Mark the block as free
  pHead->flags = MCB_FLAG_LAST & ~MCB_FLAG_ALLOCATED;

  return 1;
}

#if DONE_USED

void ALLOC_Done (tAlloc* pAlloc)
{
  if (pAlloc == NULL)
    return;
  // Free all page frames of the heap
  if (pAlloc->PageSize)
    unmap (pAlloc, pAlloc->pHeapBegin,
           (char*)pAlloc->pHeapEnd-(char*)pAlloc->pHeapBegin);
}

#endif

void* ALLOC_Alloc (tAlloc* pAlloc, size_t size)
{
  tMCBHead* pSearchHead;
  tMCBHead* pFoundHead = (tMCBHead*)NULL;

  // Check pointers, sizes and other stuff
  if ((pAlloc == NULL) || !size)
    return NULL;

  // Can't allocate blocks smaller than MIN_BLOCK_SIZE
  if (size < MIN_BLOCK_SIZE)
    size = MIN_BLOCK_SIZE;

  // Search the linked list of free blocks for a block that fits best
  for (pSearchHead = pAlloc->pHeapBegin;
       pSearchHead < (tMCBHead*)pAlloc->pHeapEnd;
       pSearchHead = (tMCBHead*)((char*)pSearchHead + pSearchHead->size + MCB_HEAD_TAIL_SIZE))
  {
    if (!(pSearchHead->flags & MCB_FLAG_ALLOCATED))
    {
      if (pSearchHead->size == size)
      {
        pFoundHead = pSearchHead;
        break;
      }

      if (pSearchHead->size > size)
      {
        if (pFoundHead == NULL)
          pFoundHead = pSearchHead;
        else
          if (pSearchHead->size < pFoundHead->size)
            pFoundHead = pSearchHead;
      }
    }
  }

  if (pFoundHead != (tMCBHead*)NULL)
  {
    // Process the found block
    tMCBHead* pHead = pFoundHead;
    tMCBTail* pTail;
    void* pBuf = (void*)((char*)pHead + sizeof(tMCBHead));
    unsigned IsLast;

    if (pHead->size <= size + MIN_MCB_SIZE)
    {
      // Allocate page frames for the block
      if (pAlloc->PageSize)
        if (!map (pAlloc, pBuf, pHead->size))
          return NULL;          // error: failed to map in page frames
      // Mark the block as allocated
      pHead->flags |= MCB_FLAG_ALLOCATED;
      // The found block can't be split any further, return it
      return pBuf;
    }

    // Allocate page frames for the block and a pair of tail and head tags
    if (pAlloc->PageSize)
      if (!map (pAlloc, pBuf, size + MIN_MCB_SIZE))
        return NULL;            // error: failed to map in page frames

    // Mark the block as allocated
    pHead->flags |= MCB_FLAG_ALLOCATED;

    IsLast = pHead->flags & MCB_FLAG_LAST;
    pHead->flags &= ~MCB_FLAG_LAST;
    // Split the found block in two: used(allocated) and unused(free)
    pTail = (void*)((char*)pHead + sizeof(tMCBHead) + pHead->size);
    pHead->size = size;
    pTail->size -= size + MCB_HEAD_TAIL_SIZE;

    pHead = (void*)((char*)pTail - (pTail->size + sizeof(tMCBHead)));
    pHead->size = pTail->size;
    pHead->flags = IsLast & ~MCB_FLAG_ALLOCATED;

    pTail = (void*)((char*)pHead - sizeof(tMCBTail));
    pTail->size = pFoundHead->size;

    // Return the allocated block
    return pBuf;
  }

  // No block was found, return NULL
  return NULL;
}

void ALLOC_Free (tAlloc* pAlloc, void* pBuf)
{
  tMCBHead* pHead = (void*)((char*)pBuf - sizeof(tMCBHead));
  tMCBTail* pTail;
  tMCBHead* pNeibHead;
  tMCBTail* pNeibTail;

  // Check the input pointer and other stuff
  if ((pAlloc == NULL) || (pBuf == NULL))
    return;

  if ((pBuf <  (void*)((char*)pAlloc->pHeapBegin + sizeof(tMCBHead))) ||
      (pBuf >= (void*)((char*)pAlloc->pHeapEnd   - sizeof(tMCBTail))))
    return;                     // error: bad pointer

  pTail = (void*)((char*)pBuf + pHead->size);

  if (pTail->size != pHead->size)
    return;                     // error: bad pointer or memory corrupted

  if (!(pHead->flags & MCB_FLAG_ALLOCATED))
    return;                     // error: this buffer isn't allocated

  // Mark the block as free
  pHead->flags &= ~MCB_FLAG_ALLOCATED;

  // Coalesce with next block
  pNeibHead = (void*)((char*)pTail + sizeof(tMCBTail));
  if ((void*)pNeibHead < (void*)((char*)pAlloc->pHeapEnd - MCB_HEAD_TAIL_SIZE))
  {
    pNeibTail = (void*)((char*)pNeibHead + sizeof(tMCBHead) + pNeibHead->size);
    if ((void*)pNeibTail <= (void*)((char*)pAlloc->pHeapEnd - sizeof(tMCBTail)))
      if (pNeibHead->size == pNeibTail->size)
        if (!(pNeibHead->flags & MCB_FLAG_ALLOCATED))
        {
          if (pNeibHead->flags & MCB_FLAG_LAST)
            pHead->flags |= MCB_FLAG_LAST;

          // Finally, coalesce the blocks
          pTail = pNeibTail;
          pHead->size += pNeibHead->size + MCB_HEAD_TAIL_SIZE;
          pTail->size = pHead->size;
        }
  }

  // Coalesce with prev block
  pNeibTail = (void*)((char*)pHead - sizeof(tMCBTail));
  if ((void*)pNeibTail > (void*)((char*)pAlloc->pHeapBegin + sizeof(tMCBHead)))
  {
    pNeibHead = (void*)((char*)pNeibTail - (pNeibTail->size + sizeof(tMCBHead)));
    if ((void*)pNeibHead >= pAlloc->pHeapBegin)
      if (pNeibHead->size == pNeibTail->size)
        if (!(pNeibHead->flags & MCB_FLAG_ALLOCATED))
        {
          if (pHead->flags & MCB_FLAG_LAST)
            pNeibHead->flags |= MCB_FLAG_LAST;

          // Finally, coalesce the blocks
          pNeibHead->size += pHead->size + MCB_HEAD_TAIL_SIZE;
          pTail->size = pNeibHead->size;
          pHead = pNeibHead;
        }
  }

  // Free page frames of the block
  if (pAlloc->PageSize)
    unmap (pAlloc, (char*)pHead + sizeof(tMCBHead), pHead->size);
}

#if STATISTICS_USED

void ALLOC_GetStatistics (tAlloc* pAlloc, tAllocStatistics* pStat)
{
  tMCBHead* pSearchHead;
  tMCBTail* pSearchTail;

  // Check the pointers and other stuff
  if (pStat != NULL)
  {
    // Initialize statistics
    pStat->Errors = 
      pStat->BlocksAllocated = pStat->BlocksFree = pStat->BlocksTotal = 0;

    pStat->TreeHeight = pStat->LargestFreeSize = 
      pStat->MemoryAllocated = pStat->MemoryFree = pStat->MemoryTotal = 0;
  }

  if ((pAlloc == NULL) || (pStat == NULL))
    return;

  // Loop through all of the blocks
  for (pSearchHead = pAlloc->pHeapBegin;
       pSearchHead < (tMCBHead*)pAlloc->pHeapEnd;
       pSearchHead = (void*)((char*)pSearchHead + pSearchHead->size + MCB_HEAD_TAIL_SIZE))
  {
    pSearchTail = (void*)((char*)pSearchHead + sizeof(tMCBHead) + pSearchHead->size);

    // Update statistics
    pStat->BlocksTotal++;
    pStat->MemoryTotal += pSearchHead->size;
    pStat->Errors += (pSearchHead->size != pSearchTail->size);

    if (pSearchHead->flags & MCB_FLAG_ALLOCATED)
    {
      pStat->BlocksAllocated++;
      pStat->MemoryAllocated += pSearchHead->size;
    }
    else
    {
      pStat->BlocksFree++;
      pStat->MemoryFree += pSearchHead->size;
      if (pSearchHead->size > pStat->LargestFreeSize)
        pStat->LargestFreeSize = pSearchHead->size;
    }
  }
}

#endif
